from hackerforms.common import Input
from typing import Union, Optional
import datetime
import time


class DateInput(Input):
    type = "date-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.initial_value = kwargs.get("initial_value", None)
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.disabled = kwargs.get("disabled", False)

    @staticmethod
    def convert_value(value: Union[datetime.date, time.struct_time, str]) -> str:
        if isinstance(value, datetime.date):
            return value.isoformat()
        elif isinstance(value, time.struct_time):
            return (
                datetime.datetime.fromtimestamp(time.mktime(value)).date().isoformat()
            )
        return value

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "label": self.label,
            "initialValue": DateInput.convert_value(self.initial_value)
            if self.initial_value
            else "",
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
        }

    @staticmethod
    def __revert_value(value: Optional[datetime.date]):
        if not value:
            return None
        return value.isoformat()

    @staticmethod
    def __convert_answer(answer: str) -> Optional[datetime.date]:
        if not answer:
            return None
        split_answer = answer.split("-")
        year = int(split_answer[0])
        month = int(split_answer[1])
        day = int(split_answer[2])
        return datetime.date(year, month, day)

    def convert_answer(self, answer: str) -> Optional[datetime.date]:
        """
        Returns:
            datetime.date: The value entered by the user
        """
        return self.__convert_answer(answer)



import os
import sys
import builtins

from .socket import send
from .page import Page

try:
    from .generated.inputs import read
    from .generated.outputs import display_plotly
except ImportError:
    pass


def writeWraper(type, write, text):
    try:
        write(text)
        send({"type": type, "log": text})
    finally:
        return len(text)


def _overload_stdio():
    stdout_write = sys.stdout.write
    stderr_write = sys.stderr.write

    sys.stdout.write = lambda text: writeWraper("stdout", stdout_write, text)
    sys.stderr.write = lambda text: writeWraper("stderr", stderr_write, text)


def _overload_input():
    read_overload = lambda msg="": read(msg)
    builtins.input = read_overload
    builtins.raw_input = read_overload


def _overload_print():
    def print_overload(*values):
        page = Page()
        for value in values:
            page.display(value)
        return page.run()

    builtins.print = print_overload


def _overload_plotly_show():
    try:
        from plotly.io import renderers
        from plotly.graph_objects import Figure
        from plotly.io._base_renderers import ExternalRenderer

        class FormsRenderer(ExternalRenderer):
            def render(self, fig_dict):
                fig = Figure(data=fig_dict)
                display_plotly(fig)
                return {}

        renderers["hackerforms"] = FormsRenderer()
        renderers.default = "hackerforms"
    except Exception as e:
        pass


def initialize():
    _overload_stdio()
    _overload_input()
    _overload_plotly_show()
    if os.getenv("ENABLE_PRINT_AS_DISPLAY") == "true":
        _overload_print()



from hackerforms.widgets.input_types import *
from hackerforms.widgets.output_types import *
from hackerforms.reactive import Reactive
from hackerforms.types import PlotlyFigure, PandasDataFrame
from typing import List, Union, Dict, Any
from hackerforms.common import Input
import io


class WidgetSchema:
    def __init__(self):
        self.widgets: List = []

    def reactive(self, callback):
        self.widgets.append(Reactive(callback))
        return self

    def convert_answer(self, form_answers: Dict) -> Dict:
        """Convert the answer from the form to the expected format
        Args:
            answer: The answer from the form
        Returns:
            The converted answer
        """
        answer: Dict = {}
        inputs = self.get_input_widgets()
        for input in inputs:
            answer[input.key] = input.convert_answer(form_answers.get(input.key))
        return answer

    def get_input_widgets(self):
        concrete_widgets = []
        for widget in self.widgets:
            if isinstance(widget, Reactive):
                concrete_widgets.extend(widget.get_widgets())
            else:
                concrete_widgets.append(widget)

        inputs = list(
            filter(lambda widget: isinstance(widget, Input), concrete_widgets)
        )
        return inputs

    def json(self, payload):
        output = []
        for widget in self.widgets:
            widget_json = widget.json(payload=payload)
            if isinstance(widget_json, list):
                output.extend(widget_json)
            else:
                output.append(widget_json)

        return output

    def read_answer_sheet(
        self, label: str, options: list, number_of_questions: int, **kwargs
    ):

        """Retrieve the answers from a test on a usual answersheet

        Position Args:
                label (str): The label to display to the user
                options (list): The options which can be chosen as an answer
                number_of_questions (int): Number of questions the answersheet will cover

        Keyword Args:
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list: The values/value selected by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(
            AnswerSheetInput(key, label, options, number_of_questions, **kwargs)
        )
        return self

    def read_cards(self, label: str, options: list, **kwargs):

        """Read a text value from the user simple text input

                Position Args:
                        label (str): The text related to this field
                        options (list): The options to display to the user, eg. [
        {'title': 'Option 1', 'subtitle': 'Subtitle 1', 'image': 'https://image_1.png', 'description': 'option 1 description', 'topLeftExtra': 'Left 1', 'topRightExtra': 'Right 1' },
        {'title': 'Option 2', 'subtitle': 'Subtitle 2', 'image': 'https://image_2.png', 'description': 'option 2 description', 'topLeftExtra': 'Left 2', 'topRightExtra': 'Right 2' }]

                Keyword Args:
                        multiple (bool): Whether the user can select multiple options. Defaults to False.
                        initial_value (list): The initial value to display to the user. Defaults to None.
                        searchable (bool): Whether to show a search bar. Defaults to False.
                        layout (str): Whether the cards layout should be 'list' or 'grid'. Defaults to 'list'.%%%
                        disabled (bool): whether the input is disabled. Defaults to False.
                        required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                        hint (str): A tooltip displayed to the user. Defaults to None.
                        end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                        full_width (bool): Whether the input should use full screen width. Defaults to False.

                Returns:
                  list, any: The options/option selected by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CardsInput(key, label, options, **kwargs))
        return self

    def read_checkbox(self, label: str, **kwargs):

        """Read a checkbox value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list(str) or list(float): The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CheckboxInput(key, label, **kwargs))
        return self

    def read_checklist(self, label: str, options: list, **kwargs):

        """Read a checklist value from the user

        Position Args:
                label (str): The label to display to the user
                options (list): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list or any: The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(ChecklistInput(key, label, options, **kwargs))
        return self

    def read_cnpj(self, label: str, **kwargs):

        """Read a CNPJ value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to None.
                placeholder (str): The placeholder text to display to the user. Defaults to "00.000.000/0001-00".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list(str) or list(float): The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CnpjInput(key, label, **kwargs))
        return self

    def read_code(self, label: str, **kwargs):

        """Read a piece of code from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                language (str): The programming language. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CodeInput(key, label, **kwargs))
        return self

    def read_cpf(self, label: str, **kwargs):

        """Read a CPF value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                placeholder (str): The placeholder text to display to the user. Defaults to "000.000.000-00".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CpfInput(key, label, **kwargs))
        return self

    def read_currency(self, label: str, **kwargs):

        """Read currency value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to 0.
                placeholder (str): The placeholder text to display to the user. Defaults to "Your answer here".
                min (float): The minimum value allowed, eg. "0". Defaults to None.
                max (float): The maximum value allowed, eg. "100". Defaults to None.
                step (float): The value to be incremented or decremented while using the input button. Defaults to None.
                currency (str): The currency to display to the user, eg. "USD", "BRL, "EUR", "GBP". Defaults to "USD".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(CurrencyInput(key, label, **kwargs))
        return self

    def read_date(self, label: str, **kwargs):

        """Read a date value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (Union[datetime.date, time.struct_time, str]): The initial value to display to the user. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(DateInput(key, label, **kwargs))
        return self

    def read_dropdown(self, label: str, options: list, **kwargs):

        """Read a dropdown value from the user

        Position Args:
                label (str): The label to display to the user
                options (list): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]

        Keyword Args:
                multiple (bool): Whether the user can select multiple options. Defaults to False.
                initial_value (str): The initial value to display to the user. Defaults to None.
                placeholder (str): The placeholder text to display to the user. Defaults to "Choose an option".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value selected by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(DropdownInput(key, label, options, **kwargs))
        return self

    def read_email(self, label: str, **kwargs):

        """Read an email value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                placeholder (str): The placeholder text to display to the user. Defaults to "Your email here".
                invalid_email_message (str): Invalid e-mail message
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(EmailInput(key, label, **kwargs))
        return self

    def read_file(self, label: str, **kwargs):

        """Read a file value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
                max_file_size (float): Maximum size allowed to be transfered in total in MB.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          A dict containing the file uploaded by the user ({"file": file, "url": str, "content": bytes}) or a list of files in case of multiple flag set as True
        """

        key = kwargs.pop("key", label)

        self.widgets.append(FileInput(key, label, **kwargs))
        return self

    def read_image(self, label: str, **kwargs):

        """Read a image file value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
                max_file_size (float): Maximum size allowed to be transfered in total in MB.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          A dict containing the image file uploaded by the user ({"file": file, "url": str, "content": bytes}) or a list of images in case of multiple flag set as True
        """

        key = kwargs.pop("key", label)

        self.widgets.append(ImageInput(key, label, **kwargs))
        return self

    def read_list(self, item_schema: Any, **kwargs):

        """Read a list value from the user

        Position Args:
                item_schema (Any): The schema for the items of the list

        Keyword Args:
                initial_value (Any): The initial value to display to the user. Defaults to [{}].
                hint (str): A tooltip displayed to the user. Defaults to None.
                min (float): Min value accepted by the input. Defaults to None.
                max (float): Max value accepted by the input. Defaults to None.
                add_button_text (str): Label to be displayed on the add button. Defaults to "+".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The values entered by the user
        """

        key = kwargs.pop("key", "result")

        self.widgets.append(ListInput(key, item_schema, **kwargs))
        return self

    def read_multiple_choice(self, label: str, options: list, **kwargs):

        """Read a multiple choice value from the user

        Position Args:
                label (str): The label to display to the user
                options (list): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]

        Keyword Args:
                multiple (bool): Whether the user can select multiple options. Defaults to False.
                min (number): The minimal amount of options that should be selected. Defaults to None.
                max (number): The maximum amount of options that should be selected. Defaults to None.
                initial_value (str): The initial value to display to the user. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list or any: The values/value selected by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(MultipleChoiceInput(key, label, options, **kwargs))
        return self

    def read_nps(self, label: str, **kwargs):

        """Gets NPS feedback from user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                min (int): Min value accepted by the input. Defaults to 0.
                max (int): Max value accepted by the input. Defaults to 10.
                min_hint (str): Text to display next to the min value. Defaults to "Not at all likely".
                max_hint (str): Text to display next to the max value. Defaults to "Extremely likely".
                initial_value (str): The initial value to display to the user. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(NpsInput(key, label, **kwargs))
        return self

    def read_number(self, label: str, **kwargs):

        """Read a number value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to 0.
                placeholder (str): The placeholder text to display to the user. Defaults to "Your answer here".
                min (float): Min value accepted by the input. Defaults to None.
                max (float): Max value accepted by the input. Defaults to None.
                step (float): The value to be incremented or decremented while using the input button. Defaults to None.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(NumberInput(key, label, **kwargs))
        return self

    def read_number_slider(self, label: str, **kwargs):

        """Read a number value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (float): The initial value to display to the user. Defaults to 0.
                min (float): Min value accepted by the input. Defaults to None.
                max (float): Max value accepted by the input. Defaults to None.
                step (float): The value to be incremented or decremented while using the input button. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(NumberSliderInput(key, label, **kwargs))
        return self

    def read_pandas_row_selection(self, df: PandasDataFrame, **kwargs):

        """Display a pandas dataframe as a table and allow the user to select rows

        Position Args:
                df (PandasDataFrame): The pandas dataframe to be displayed

        Keyword Args:
                display_index (bool): Whether to show a index column. Defaults to False.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The list of selected rows
        """

        key = kwargs.pop("key", "result")

        self.widgets.append(PandasRowSelectionInput(key, df, **kwargs))
        return self

    def read_password(self, label: str, **kwargs):

        """Read a password value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                placeholder (str): The placeholder text to display to the user. Defaults to "".
                lowercase_required (bool): Whether the input must have at least one lowercase character. Defaults to True.
                uppercase_required (bool): Whether the input must have at least one uppercase character. Defaults to True.
                special_required (bool): Whether the input must have at least one special character. Defaults to True.
                digit_required (bool): Whether the input must have at least one digit. Defaults to True.
                min_length (int): Minimum length of the password. Defaults to 8.
                max_length (int): Maximum length of the password. Defaults to None.
                size (int): Size of the password. Defaults to None.
                pattern (str): A regex pattern for the accepted password. Defaults to None.
                autocomplete (str): The autocomplete HTML attribute. Defaults to "current-password".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(PasswordInput(key, label, **kwargs))
        return self

    def read_phone(self, label: str, **kwargs):

        """Read a phone value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (Union[str, dict]): The initial value to display to the user. If dictionary, it contains two keys: 'country_code' (string with optional + sign or number) and 'national_number' (str or number). Ex: {'country_code': '+55', 'national_number': '21999990000'}. Defaults to "".
                placeholder (str): The placeholder text to display to the user. Defaults to "".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          A dict containing the value entered by the user ({"raw": str, "masked": str})
        """

        key = kwargs.pop("key", label)

        self.widgets.append(PhoneInput(key, label, **kwargs))
        return self

    def read_rating(self, label: str, **kwargs):

        """Read a rating value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (int): The initial value to display to the user. Defaults to 0.
                max (float): Max value accepted by the input. Defaults to None.
                char (str): Which char should be displayed as icon?
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(RatingInput(key, label, **kwargs))
        return self

    def read_tag(self, label: str, **kwargs):

        """Read a tag value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (list): The initial value to display to the user. Defaults to [].
                placeholder (str): The placeholder text to display to the user. Defaults to "Your answer here".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          list(str) or list(float): The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(TagInput(key, label, **kwargs))
        return self

    def read(self, label: str, **kwargs):

        """Read a text value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                placeholder (str): The placeholder text to display to the user. Defaults to "Your answer here".
                mask (str): A mask to apply to the input. Defaults to None.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(TextInput(key, label, **kwargs))
        return self

    def read_textarea(self, label: str, **kwargs):

        """Read a textarea value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                placeholder (str): The placeholder text to display to the user. Defaults to "Your answer here".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          The value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(TextareaInput(key, label, **kwargs))
        return self

    def read_time(self, label: str, **kwargs):

        """Read a time value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (str): The initial value to display to the user. Defaults to "".
                format (str): Whether the input is in the format 24hs or AM/PM. Defaults to "24hs".
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          A datetime.time object representing the value entered by the user
        """

        key = kwargs.pop("key", label)

        self.widgets.append(TimeInput(key, label, **kwargs))
        return self

    def read_toggle(self, label: str, **kwargs):

        """Read a toggle value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                on_text (str): Text of On Toggle option
                off_text (str): Text of Off Toggle option
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          bool: if the toggle was checked
        """

        key = kwargs.pop("key", label)

        self.widgets.append(ToggleInput(key, label, **kwargs))
        return self

    def read_video(self, label: str, **kwargs):

        """Read a video file value from the user

        Position Args:
                label (str): The label to display to the user

        Keyword Args:
                initial_value (Union[str, io.IOBase]): The initial value to display to the user. Defaults to "".
                multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
                max_file_size (float): Maximum size allowed to be transfered in total in MB.
                disabled (bool): whether the input is disabled. Defaults to False.
                required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
                hint (str): A tooltip displayed to the user. Defaults to None.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.

        Returns:
          A dict containing the video uploaded by the user ({"file": file, "url": str, "content": bytes}) or a list of videos in case of multiple flag set as True
        """

        key = kwargs.pop("key", label)

        self.widgets.append(VideoInput(key, label, **kwargs))
        return self

    def display_file(self, file: Union[str, io.IOBase], **kwargs):

        """Display a button for the user to download a file

        Position Args:
                file (Union[str, io.IOBase]): The file to download

        Keyword Args:
                download_text (str): The text to display on the button that will download the file. Defaults to "Download here".
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(FileOutput(file, **kwargs))
        return self

    def display_html(self, html: str, **kwargs):

        """Display a html snippet to the user

        Position Args:
                html (str): The html snippet to display to the user

        Keyword Args:
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(HtmlOutput(html, **kwargs))
        return self

    def display_iframe(self, url_or_html: str, **kwargs):

        """Display an inline iframe to the user

        Position Args:
                url_or_html (str): The link to the document or the own document to display to the user

        Keyword Args:
                width (int): The width of the iframe. Defaults to "800".
                height (int): The height of the iframe. Defaults to "600".
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(IframeOutput(url_or_html, **kwargs))
        return self

    def display_image(self, image: Union[str, io.IOBase], **kwargs):

        """Display an image to the user

        Position Args:
                image (Union[str, io.IOBase]): The image to display to the user

        Keyword Args:
                subtitle (str): The subtitle of the image. Defaults to "".
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(ImageOutput(image, **kwargs))
        return self

    def display_latex(self, text: str, **kwargs):

        """Display a latex formula to the user

        Position Args:
                text (str): The latex formula to display to the user

        Keyword Args:
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(LatexOutput(text, **kwargs))
        return self

    def display_link(self, link_url: str, **kwargs):

        """Display a link to the user

        Position Args:
                link_url (str): The url of the link to display to the user

        Keyword Args:
                link_text (str): The text to display on the link. Defaults to "Click here".
                same_tab (bool): Whether to open the link in the same tab or not. Defaults to False.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(LinkOutput(link_url, **kwargs))
        return self

    def display_markdown(self, text: str, **kwargs):

        """Display a formatted text to the user

        Position Args:
                text (str): The formatted text to display to the user

        Keyword Args:
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(MarkdownOutput(text, **kwargs))
        return self

    def display_pandas(self, df: PandasDataFrame, **kwargs):

        """Display a pandas dataframe to the user

        Position Args:
                df (PandasDataFrame): The dataframe to display to the user

        Keyword Args:
                display_index (bool): Whether to show a index column. Defaults to False.
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(PandasOutput(df, **kwargs))
        return self

    def display_plotly(self, fig: PlotlyFigure, **kwargs):

        """Display a plotly figure to the user

        Position Args:
                fig (PlotlyFigure): The figure to display to the user

        Keyword Args:
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(PlotlyOutput(fig, **kwargs))
        return self

    def display_progress(self, current: float, total: float, **kwargs):

        """Display a progress bar. This widget is shown on screen until the script shows a new widget. This widget does not work on Pages, and must be used only with its function.

        Position Args:
                current (float): The progress being made. Defaults to 50.
                total (float): Total progress. Defaults to 100.

        Keyword Args:
                text (str): The text displayed with this progress step. Defaults to "".
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(ProgressOutput(current, total, **kwargs))
        return self

    def display(self, text: str, **kwargs):

        """Display a text to the user

        Position Args:
                text (str): The text to display to the user

        Keyword Args:
                end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
                full_width (bool): Whether the input should use full screen width. Defaults to False.


        """

        self.widgets.append(TextOutput(text, **kwargs))
        return self

    input = read



from hackerforms.common import Input


class PasswordInput(Input):
    type = "password-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.hint = kwargs.get("hint", None)
        self.required = kwargs.get("required", True)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.lowercase_required = kwargs.get("lowercase_required", True)
        self.uppercase_required = kwargs.get("uppercase_required", True)
        self.special_required = kwargs.get("special_required", True)
        self.digit_required = kwargs.get("digit_required", True)
        self.min_length = kwargs.get("min_length", 8)
        self.max_length = kwargs.get("max_length", None)
        self.size = kwargs.get("size", None)
        self.pattern = kwargs.get("pattern", None)
        self.autocomplete = kwargs.get("autocomplete", "current-password")
        self.placeholder = kwargs.get("placeholder", "")
        self.disabled = kwargs.get("disabled", False)
        self.secret = True

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "hint": self.hint,
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "lowercaseRequired": self.lowercase_required,
            "uppercaseRequired": self.uppercase_required,
            "specialRequired": self.special_required,
            "digitRequired": self.digit_required,
            "minLength": self.min_length,
            "maxLength": self.max_length,
            "size": self.size,
            "pattern": self.pattern,
            "autocomplete": self.autocomplete,
            "placeholder": self.placeholder,
            "disabled": self.disabled,
            "secret": self.secret,
        }

    def convert_answer(self, answer: str) -> str:
        """
        Returns:
            str: The value entered by the user
        """
        return answer



from hackerforms.widgets.TextareaInput import TextareaInput
from hackerforms.widgets.ListInput import ListInput
from hackerforms.widgets.VideoInput import VideoInput
from hackerforms.widgets.TagInput import TagInput
from hackerforms.widgets.TimeInput import TimeInput
from hackerforms.widgets.NumberSliderInput import NumberSliderInput
from hackerforms.widgets.CheckboxInput import CheckboxInput
from hackerforms.widgets.CnpjInput import CnpjInput
from hackerforms.widgets.CpfInput import CpfInput
from hackerforms.widgets.TextInput import TextInput
from hackerforms.widgets.PhoneInput import PhoneInput
from hackerforms.widgets.ImageInput import ImageInput
from hackerforms.widgets.ToggleInput import ToggleInput
from hackerforms.widgets.ChecklistInput import ChecklistInput
from hackerforms.widgets.EmailInput import EmailInput
from hackerforms.widgets.MultipleChoiceInput import MultipleChoiceInput
from hackerforms.widgets.CodeInput import CodeInput
from hackerforms.widgets.DateInput import DateInput
from hackerforms.widgets.PasswordInput import PasswordInput
from hackerforms.widgets.NumberInput import NumberInput
from hackerforms.widgets.CardsInput import CardsInput
from hackerforms.widgets.AnswerSheetInput import AnswerSheetInput
from hackerforms.widgets.NpsInput import NpsInput
from hackerforms.widgets.CurrencyInput import CurrencyInput
from hackerforms.widgets.FileInput import FileInput
from hackerforms.widgets.RatingInput import RatingInput
from hackerforms.widgets.ClickInput import ClickInput
from hackerforms.widgets.DropdownInput import DropdownInput
from hackerforms.widgets.PandasRowSelectionInput import PandasRowSelectionInput



from tempfile import NamedTemporaryFile
import requests
from dataclasses import dataclass


class FileResponse:
    """A file response from the user

    Attributes:
        file (file): The file object
        url (str): The url of the file
        content (bytes): The content of the file
    """

    def __init__(self, url):
        res = requests.get(url)
        self.content = res.content
        self.url = url
        self.name = url.split("/")[-1]

        self.file = NamedTemporaryFile()
        self.file.write(self.content)
        self.file.seek(0)


@dataclass
class PhoneResponse:
    """A phone response from the user

    Attributes:
        masked (str): The masked phone number, eg: +55 (21) 99999-9999
        raw (str): The raw phone number, eg: 5521999999999
        country_code (str): The phone number country code, eg: 55
        national_number (str): The phone number national number, eg: 21999999999
    """

    masked: str
    raw: str
    country_code: str
    national_number: str



from hackerforms.common import Input
from typing import List, Any
import json


class PandasRowSelectionInput(Input):
    type = "pandas-row-selection-input"

    def __init__(self, key: str, df: Any, **kwargs):
        super().__init__(key)
        self.df = df
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.display_index = kwargs.get("display_index", False)
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "table": json.loads(self.df.to_json(orient="table")),
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "displayIndex": self.display_index,
            "disabled": self.disabled,
        }

    def convert_answer(self, answer) -> List:
        """
        Returns:
            list: The list of selected rows
        """
        return answer



import os, inspect, typing

from ..debug_utils import CloseDTO, Frames, make_debug_data
from ..utils import serialize, deserialize, persist_session_id, open_browser


WS_HOST = os.environ.get("WS_HOST", "wss://hackerforms-broker.abstra.cloud")
FRONTEND_HOST = os.environ.get("FRONTEND_HOST", "https://console.abstracloud.com")


class Connection:
    session_id = None
    debug_enabled = False
    ws = None
    url_params: dict = {}

    def __init__(
        self,
        session_id: typing.Union[str, None] = None,
        debug_enabled: bool = False,
    ) -> None:
        from websocket import create_connection

        self.session_id = session_id
        self.debug_enabled = debug_enabled

        if session_id:
            self.ws = create_connection(f"{WS_HOST}/lib?sessionId={session_id}")
        else:
            self.ws = create_connection(f"{WS_HOST}/lib")
            self.session_id = self.receive("sessionId")
            open_browser(FRONTEND_HOST, self.session_id)
            persist_session_id(self.session_id)

        start = {"type": None}
        while start["type"] != "start":
            start = self.receive()
        self.url_params = start["params"]

    def close(self, dto: CloseDTO):
        try:
            if self.ws is None or not self.ws.connected:
                return
            self.send(
                {
                    "type": "program:end",
                    "exitCode": dto.exit_code,
                    "exception": dto.exception,
                },
                frames=dto.frames,
            )
            self.ws.close()
        except:
            pass

    def send(self, data, frames: Frames = None):
        if self.debug_enabled:
            debug = make_debug_data(frames or inspect.stack())
            self.ws.send(serialize({**data, **debug}))
        else:
            self.ws.send(serialize(data))

    def receive(self, path: str = ""):
        recvd = self.ws.recv()
        if not recvd:
            raise Exception("Websocket not connected")
        data = deserialize(recvd)
        if not path:
            return data
        return data.get(path, None)



from hackerforms.common import Output


class LatexOutput(Output):
    type = "latex-output"

    def __init__(self, text: str, **kwargs):
        self.text = text
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "text": self.text,
            "columns": self.columns,
            "fullWidth": self.full_width,
        }



from hackerforms.widgets.PandasOutput import PandasOutput
from hackerforms.widgets.ProgressOutput import ProgressOutput
from hackerforms.widgets.LinkOutput import LinkOutput
from hackerforms.widgets.ImageOutput import ImageOutput
from hackerforms.widgets.HtmlOutput import HtmlOutput
from hackerforms.widgets.FileOutput import FileOutput
from hackerforms.widgets.MarkdownOutput import MarkdownOutput
from hackerforms.widgets.PlotlyOutput import PlotlyOutput
from hackerforms.widgets.IframeOutput import IframeOutput
from hackerforms.widgets.LatexOutput import LatexOutput
from hackerforms.widgets.TextOutput import TextOutput



import json
import requests
import urllib.request
import urllib.response


def upload_file(file):
    response = requests.post(
        "https://upload.abstra.cloud/hackerforms/upload",
        data=json.dumps({"filepath": file.name}),
        headers={"content-type": "application/json"},
    )
    response_json = json.loads(response.text)

    file.seek(0)
    content = file.read()
    content = str.encode(content) if not isinstance(content, bytes) else content

    req = urllib.request.Request(
        url=response_json["putURL"], method="PUT", data=content
    )
    urllib.request.urlopen(req)

    return response_json["getURL"]



from abc import abstractmethod, ABC


class Input(ABC):
    type: str

    def __init__(self, key: str) -> None:
        super().__init__()
        self.key = key

    @abstractmethod
    def json(self, **kwargs):
        pass

    @abstractmethod
    def convert_answer(self, answer):
        pass


class Output(ABC):
    type: str

    @abstractmethod
    def json(self, **kwargs):
        pass



from functools import reduce
from typing import List, Union, Callable
from .page import Page
from .page_response import PageResponse

Step = Union[Page, Callable]
Steps = List[Step]


class StepsResponse:
    def __init__(self):
        self.steps: Steps = []
        self.acc = {}

    def __getitem__(self, key: Union[int, str]):
        if isinstance(key, int):
            return self.steps[key]
        if isinstance(key, str):
            return self.acc[key]
        raise IndexError("Step not found")

    def append(self, step: Step):
        self.steps.append(step)
        self.__update_acc()

    def __update_acc(self):
        self.acc = {}
        for step in self.steps:
            if isinstance(step, PageResponse):
                self.acc.update(step)

    def pop(self, i: int = -1):
        step = self.steps.pop(i)
        self.__update_acc()
        return step


def run_steps(steps: Steps) -> StepsResponse:
    steps_to_execute: Steps = []  # stack
    executed_steps: Steps = []  # stack
    responses: StepsResponse = StepsResponse()  # stack

    steps_to_execute = steps[::-1]
    if not steps_to_execute:
        return responses

    if isinstance(steps_to_execute[-1], Callable):
        raise ValueError("First step cannot be a function")

    if steps_to_execute:
        response = run_page(
            steps_to_execute, executed_steps, responses, first_page=True
        )

    while steps_to_execute or response.action == "Back":
        if steps_to_execute and response.action != "Back":
            responses = execute_functions(steps_to_execute, executed_steps, responses)
            if not steps_to_execute:
                break
            response = run_page(
                steps_to_execute, executed_steps, responses, first_page=False
            )

        if response.action == "Back":
            response = go_back(steps_to_execute, executed_steps, responses)

    return responses


def go_back(steps_to_execute: Steps, executed_steps: Steps, responses: StepsResponse):
    page = executed_steps.pop()
    responses.pop()
    steps_to_execute.append(page)

    undo_functions(steps_to_execute, executed_steps, responses)
    if len(executed_steps) == 1:
        response = run_back_page(
            steps_to_execute, executed_steps, responses, first_page=True
        )
        return response

    response = run_back_page(
        steps_to_execute, executed_steps, responses, first_page=False
    )
    return response


def execute_functions(
    steps_to_execute: Steps, executed_steps: Steps, responses: StepsResponse
) -> StepsResponse:
    while steps_to_execute and isinstance(steps_to_execute[-1], Callable):
        fn = steps_to_execute.pop()
        responses.append(fn(responses))
        executed_steps.append(fn)
    return responses


def undo_functions(
    steps_to_execute: Steps, executed_steps: Steps, responses: StepsResponse
) -> StepsResponse:
    while executed_steps and isinstance(executed_steps[-1], Callable):
        fn = executed_steps.pop()
        steps_to_execute.append(fn)
        responses.pop()


def run_page(
    steps_to_execute: Steps,
    executed_steps: Steps,
    responses: StepsResponse,
    first_page: bool = False,
):
    page = steps_to_execute.pop()
    response = page.run() if first_page else page.run(actions=["Back", "Next"])
    responses.append(response)
    executed_steps.append(page)
    return response


def run_back_page(
    steps_to_execute: Steps,
    executed_steps: Steps,
    responses: StepsResponse,
    first_page: bool = False,
):
    page = executed_steps.pop()
    old_response: PageResponse = responses.pop()
    response = (
        page.run(initial_payload=old_response)
        if first_page
        else page.run(actions=["Back", "Next"], initial_payload=old_response)
    )
    responses.append(response)
    executed_steps.append(page)
    return response



import os


def check_version():
    if os.environ.get("SESSION_ID"):
        return

    import pkg_resources
    import requests

    __version__ = pkg_resources.get_distribution("hackerforms").version

    try:
        libs = requests.get(
            "https://hackerforms-api.abstra.cloud/public/abstra-pypi-packages"
        ).json()
        hackerforms = list(filter(lambda lib: lib["name"] == "hackerforms", libs))[0]
        if hackerforms["version"] != __version__ and __version__ != "0.0.0":
            print("You are using an outdated version of hackerforms.")
            print(
                "Please update your library using the following command: pip install hackerforms=="
                + hackerforms["version"]
            )
    except Exception as e:
        pass



import inspect
import hackerforms.widgets as hw
from hackerforms.common import Input, Output


def get_widget_class(type: str):
    return [
        w
        for w in [
            *hw.input_types.__dict__.values(),
            *hw.output_types.__dict__.values(),
        ]
        if inspect.isclass(w)
        and (issubclass(w, Input) or issubclass(w, Output))
        and w.type == type
    ][0]



from hackerforms.common import Input
from typing import Any, List


class ListInput(Input):
    type = "list-input"
    instances = []
    instances = []

    def __init__(self, key: str, item_schema: Any, **kwargs):
        super().__init__(key)
        self.item_schema = item_schema
        self.initial_value = kwargs.get("initial_value", [{}])
        self.min = kwargs.get("min", None)
        self.max = kwargs.get("max", None)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.add_button_text = kwargs.get("add_button_text", "+")
        self.full_width = kwargs.get("full_width", False)
        self.required = kwargs.get("required", True)
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        json = {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "itemSchema": self.item_schema.json(
                payload=self.item_schema.convert_answer({})
            ),
            "initialValue": self.initial_value,
            "columns": self.columns,
            "min": self.min,
            "max": self.max,
            "addButtonText": self.add_button_text,
            "fullWidth": self.full_width,
            "required": self.required,
            "disabled": self.disabled,
        }
        overloaded_schemas = self.__get_overloaded_schemas(
            kwargs.get("payload").get(self.key) if kwargs.get("payload") else None
        )
        if overloaded_schemas:
            json["overloadedItemSchemas"] = overloaded_schemas
        return json

    def convert_answer(self, answers) -> List:
        """
        Returns:
            list: The values entered by the user
        """
        return [
            (
                self.instances[index].convert_answer(answer)
                if index < len(self.instances)
                else self.item_schema.convert_answer(answer)
            )
            for index, answer in enumerate(answers or [])
        ]

    def __get_overloaded_schemas(self, payload):
        if payload:
            self.instances = [self.item_schema.copy() for _ in payload]
            return [
                self.instances[index].json(payload=payload_item)
                for index, payload_item in enumerate(payload)
            ]
        else:
            return None



import os, atexit
from collections import UserDict
from .debug_utils import CloseDTO, Frames
from .connection import Connection
from .exit_hook import ExitHooks


WS_HOST = os.environ.get("WS_HOST", "wss://hackerforms-broker.abstra.cloud")
FRONTEND_HOST = os.environ.get("FRONTEND_HOST", "https://console.abstracloud.com")
DEBUG_ENABLED = os.environ.get("ABSTRA_DEBUG")
SESSION_ID = os.environ.get("SESSION_ID")
ABSTRA_FORM_SERVER = os.environ.get("ABSTRA_FORM_SERVER")


def initialize():
    global __connection, initialized
    if ABSTRA_FORM_SERVER:
        return
    initialized = True
    hook = ExitHooks()
    atexit.register(lambda: close(hook.close_dto))
    __connection = Connection(SESSION_ID, DEBUG_ENABLED)


def get_connection():
    if not initialized or not __connection:
        raise Exception("No connection available")
    return __connection


def send(data, frames: Frames = None):
    return get_connection().send(data, frames)


def receive(path: str = ""):
    return get_connection().receive(path)


def close(dto: CloseDTO):
    return get_connection().close(dto)


# this is a class to wrap all read methods of dicts to call get_url_params instead
class UrlParams(UserDict):
    @property
    def params(self):
        conn = get_connection()
        return conn.url_params if conn else {}

    def __getitem__(self, key):
        return self.params.__getitem__(key)

    def __iter__(self):
        return self.params.__iter__()

    def __len__(self) -> int:
        return self.params.__len__()

    def __contains__(self, key):
        return self.params.__contains__(key)

    def __str__(self) -> str:
        return self.params.__str__()

    def __repr__(self) -> str:
        return self.params.__repr__()

    def keys(self):
        return self.params.keys()

    def values(self):
        return self.params.values()

    def items(self):
        return self.params.items()

    def has_key(self, key):
        return key in self.params


url_params = UrlParams()
initialized = False
__connection = None



import os
import json
import webbrowser

from typing import Dict


def deserialize(string: str) -> Dict:
    return json.loads(string)


def serialize(obj: Dict) -> bytes:
    return bytes(json.dumps(obj, allow_nan=False), "UTF-8")


def persist_session_id(session_id: str):
    if os.getenv("ENV") == "e2e" and os.getenv("E2EPATH"):
        with open(f'{os.getenv("E2EPATH")}/session-id.txt', "w") as f:
            f.write(session_id)


def open_browser(frontend_host, session_id):
    if os.getenv("ENV") != "e2e":
        url = f"{frontend_host}/local/{session_id}"
        if not webbrowser.open(url):
            print(f"Open your browser at: {url}")


def get_single_value(answer: Dict):
    return list(answer.values())[0]



from hackerforms.common import Output
from typing import Any
import json


class PlotlyOutput(Output):
    type = "plotly-output"

    def __init__(self, fig: Any, **kwargs):
        self.fig = fig
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "figure": json.loads(self.fig.to_json()),
            "columns": self.columns,
            "fullWidth": self.full_width,
        }



from hackerforms.socket import send, receive
from hackerforms.generated.widget_schema import WidgetSchema
from hackerforms.page_response import PageResponse
from hackerforms.validation import validate_widget_props
from .reactive import Reactive

from typing import Callable, Dict, Union, List


class Page(WidgetSchema):
    """A form page that can be displayed to the user

    This is a page that can be displayed to the user. It can be used to
    show data as well as collect informations. After configuring the
    inputs and outputs, use the run method to display the form to the
    user and collect the answers.
    """

    def __init__(
        self,
        actions: Union[str, List[str]] = "Next",
        columns: float = 1,
        validate: Callable = None,
        end_program: bool = False,
        reactive_polling_interval=0,
        initial_payload: Dict = None,
    ):
        super().__init__()
        self.__actions = actions
        self.__columns = columns
        self.__validate = validate
        self.__end_program = end_program
        self.__reactive_polling_interval = reactive_polling_interval
        self.__initial_payload = initial_payload

    def run(
        self,
        actions: Union[str, List[str]] = "Next",
        columns: float = 1,
        validate: Callable = None,
        end_program: bool = False,
        reactive_polling_interval=0,
        initial_payload: Dict = None,
    ) -> Dict:
        """Run the form

        Args:
            button_text: The text of the button that is used to submit the form
            columns: The number of columns of the form
            end_program: End program whilst showing this page, making this an end page
            reactive_polling_interval: Interval to poll and rerender reactive slots, in seconds (0 to disable)

        Returns:
            The form result as a dict with the keys being the key of the input and the value being the value of the input
        """

        actions = actions if actions != "Next" else self.__actions
        columns = columns if columns != 1 else self.__columns
        end_program = end_program if end_program != False else self.__end_program
        reactive_polling_interval = (
            reactive_polling_interval
            if reactive_polling_interval != 0
            else self.__reactive_polling_interval
        )
        initial_payload = (
            initial_payload if initial_payload != None else self.__initial_payload
        )
        validate = validate if validate != None else self.__validate

        if initial_payload:
            for widget in self.widgets:
                if not isinstance(widget, Reactive) and widget.key in initial_payload:
                    widget.initial_value = initial_payload[widget.key]

        widgets_json = self.__get_validated_page_widgets_json(self.convert_answer({}))

        if self.__is_progress_screen():
            self.__send_form_message(
                widgets=widgets_json,
                columns=columns,
                actions=[],
                end_program=end_program,
                reactive_polling_interval=reactive_polling_interval,
            )
            return

        self.__send_form_message(
            widgets=widgets_json,
            columns=columns,
            actions=self.__actions_property(actions, end_program),
            end_program=end_program,
            reactive_polling_interval=reactive_polling_interval,
        )
        response: Dict = self.__user_event_messages(validate=validate)

        if end_program:
            exit()

        return PageResponse(
            self.convert_answer(response["payload"]),
            response.get("action"),
        )

    def __user_event_messages(self, **kwargs):
        response: Dict = receive()

        while response["type"] == "user-event":
            converted_payload = self.convert_answer(response["payload"])
            widgets_json = self.__get_validated_page_widgets_json(converted_payload)
            self.__send_user_event_message(
                widgets=widgets_json,
                validation=self.__build_validation_object(
                    validation=kwargs.get("validate"), payload=converted_payload
                ),
            )

            response = receive()

        return response

    def __get_validated_page_widgets_json(self, converted_payload):
        widgets_json = self.json(converted_payload)
        for widget in widgets_json:
            validate_widget_props(widget)
        return widgets_json

    def __actions_property(self, actions, end_page):
        if end_page:
            return []
        elif isinstance(actions, list):
            return actions
        elif actions is None:
            return []
        return [actions]

    def __is_progress_screen(self):
        return len(self.widgets) == 1 and self.widgets[0].type == "progress-output"

    def __build_validation_object(self, validation, payload):
        validation_status = True
        validation_message = ""

        if validation:
            validation_response = validation(payload)
            if type(validation_response) == bool:
                validation_status = validation_response
                validation_message = ""
            elif type(validation_response) == str:
                validation_status = False
                validation_message = validation_response

        return {"status": validation_status, "message": validation_message}

    def __send_form_message(
        self, widgets, actions, columns, end_program, reactive_polling_interval
    ):
        send(
            {
                "type": "form",
                "widgets": widgets,
                "columns": columns,
                "actions": actions,
                "endProgram": end_program,
                "reactivePollingInterval": reactive_polling_interval,
            }
        )

    def __send_user_event_message(self, widgets, validation):
        send(
            {
                "type": "user-event",
                "widgets": widgets,
                "validation": validation,
            }
        )



import sys, os
from .debug_utils import CloseDTO, traceback_to_infos


class ExitHooks(object):
    def __init__(self):
        self.exit_code = None
        self.exception = None
        self.frames = None

        self._orig_exit = sys.exit
        self._orig_excepthook = sys.excepthook
        sys.exit = self.exit
        sys.excepthook = self.excepthook

    @property
    def close_dto(self):
        return CloseDTO(
            exit_code=self.exit_code, exception=self.exception, frames=self.frames
        )

    def exit(self, code=0):
        self.exit_code = code
        self._orig_exit(code)

    def excepthook(self, exc_type, exc, traceback):
        self.exception = str(exc)
        if os.environ.get("ABSTRA_DEBUG"):
            self.frames = traceback_to_infos(traceback)
        self._orig_excepthook(exc_type, exc, traceback)



from hackerforms.common import Output


class LinkOutput(Output):
    type = "link-output"

    def __init__(self, link_url: str, **kwargs):
        self.link_url = link_url
        self.link_text = kwargs.get("link_text", "Click here")
        self.same_tab = kwargs.get("same_tab", False)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "linkText": self.link_text,
            "linkUrl": self.link_url,
            "columns": self.columns,
            "sameTab": self.same_tab,
            "fullWidth": self.full_width,
        }



from hackerforms.common import Input
from hackerforms.file_utils import convert_file
from typing import Union, List, Any


class CardsInput(Input):
    type = "cards-input"

    def __init__(self, key: str, label: str, options: Any, **kwargs):
        super().__init__(key)
        self.label = label
        self.options = options
        self.multiple = kwargs.get("multiple", False)
        self.searchable = kwargs.get("searchable", False)
        self.initial_value = kwargs.get("initial_value", None)
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.layout = kwargs.get("layout", "list")
        self.disabled = kwargs.get("disabled", False)

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "hint": self.hint,
            "options": [
                {**opt, "image": convert_file(opt.get("image"))} for opt in self.options
            ],
            "multiple": self.multiple,
            "searchable": self.searchable,
            "initialValue": self.initial_value,
            "required": self.required,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "layout": self.layout,
            "disabled": self.disabled,
        }

    def convert_answer(self, answer: Union[List, Any]) -> Union[List, Any]:
        """
        Returns:
            list, any: The options/option selected by the user
        """
        return answer



from .socket import send, receive


def get_user():
    """Get the users email and verify it

    This method will request users to fill in their email addresses and
    verify them. It will then return the email address once it has been validated.

    Returns:
      AuthReponse
    """
    send({"type": "auth:initialize"})
    while True:
        auth = receive()
        if auth["type"] == "auth:validation-ended":
            return AuthResponse(auth["email"])


class AuthResponse:
    """The response from the authentication process

    Attributes:
      email (str): The email address of the user
    """

    def __init__(self, email: str):
        self.email = email



import os

from .auth import *

from .actions import execute_js, redirect
from .version import check_version as _check_version
from .socket import initialize as _initialize_socket, url_params
from .overloads import initialize as _initialize_overloads
from .list_item_schema import ListItemSchema
from .page import Page
from .step import run_steps
from . import environments as envs
from .generated.inputs import *
from .generated.outputs import *

ENV = os.getenv("ENV")
RUNNER = os.getenv("RUNNER")

if not (ENV in [envs.TEST] or RUNNER in ["dash", "hook", "job"]):
    _check_version()
    _initialize_socket()
    _initialize_overloads()

del _check_version
del _initialize_socket
del _initialize_overloads



from hackerforms.common import Input
from typing import Optional
from hackerforms.response_types import PhoneResponse


class PhoneInput(Input):
    type = "phone-input"

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.label = label
        self.initial_value = kwargs.get("initial_value", "")
        self.required = kwargs.get("required", True)
        self.hint = kwargs.get("hint", None)
        self.placeholder = kwargs.get("placeholder", "")
        self.columns = kwargs.get("columns", 1)
        self.full_width = kwargs.get("full_width", False)
        self.disabled = kwargs.get("disabled", False)

    def _initial_value_to_json(self, initial_value):
        if isinstance(initial_value, str):
            return {"countryCode": "+1", "nationalNumber": initial_value}
        return {
            "countryCode": initial_value["country_code"]
            if "country_code" in initial_value
            else "",
            "nationalNumber": initial_value["national_number"]
            if "national_number" in initial_value
            else "",
        }

    def json(self, **kwargs):
        return {
            "type": self.type,
            "key": self.key,
            "label": self.label,
            "initialValue": self._initial_value_to_json(self.initial_value),
            "placeholder": self.placeholder,
            "required": self.required,
            "hint": self.hint,
            "columns": self.columns,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
        }

    @staticmethod
    def __revert_value(value: Optional[PhoneResponse]):
        if not value:
            return None
        return {
            "raw": value.raw,
            "masked": value.masked,
            "countryCode": value.country_code,
            "nationalNumber": value.national_number,
        }

    @staticmethod
    def __convert_answer(answer) -> Optional[PhoneResponse]:
        if not answer:
            return None
        return PhoneResponse(
            masked=answer["masked"],
            raw=answer["raw"],
            country_code=answer["countryCode"],
            national_number=answer["nationalNumber"],
        )

    def convert_answer(self, answer) -> Optional[PhoneResponse]:
        """
        Returns:
            PhoneResponse: A dict containing the value entered by the user ({"raw": str, "masked": str})
        """
        return self.__convert_answer(answer)
